import random

import numpy as np
import pandas as pd
import torch
import torch.utils

from ..helper import helper
from . import adapter
from . import data_frame as df_creator
from . import reader


class HistoDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        data_frame_creator: str,
        path_to_dataset: str,
        data_frame_editor=None,
        data_frame_filters=None,
        data_readers: str = None,
        feature_readers: str = None,
        pre_transfs=None,
        da_transfs=None,
        seed: int = None,
        return_data_rows: bool = False,
    ):
        super().__init__()

        # save variables
        self.path_to_dataset = path_to_dataset
        self.return_data_rows = return_data_rows
        self.seed = seed
        self.seed_increment = 0

        # create a DataFrameCreator if str was given
        if isinstance(data_frame_creator, str):
            data_frame_creator = df_creator.CreateDFFromCSV(data_frame_creator)

        # convert or create list of data_readers
        self.data_readers = helper.create_dict_from_variable(data_readers, "data")
        for name in self.data_readers:
            if isinstance(self.data_readers[name], str):
                self.data_readers[name] = reader.ReadFromImageFile(self.data_readers[name])

        # convert or create list of feature_readers
        self.feature_readers = helper.create_dict_from_variable(feature_readers, "feature")
        for name in self.feature_readers:
            if isinstance(self.feature_readers[name], str):
                self.feature_readers[name] = reader.ReadValueFromCSV(self.feature_readers[name])

        # convert or create list of pre_transformation_adapters
        if pre_transfs is None:
            if self.seed is None:
                self.pre_transfs = adapter.Adapter(seed=random.randint(0, 99999999999))
            else:
                self.pre_transfs = adapter.Adapter(seed=self.seed * int(1e3))
        else:
            self.pre_transfs = pre_transfs

        # convert or create list of pre_transformation_adapters
        if da_transfs is None:
            self.da_transfs = adapter.Adapter()
        else:
            self.da_transfs = da_transfs

        # read csv file
        self.df = data_frame_creator(self.path_to_dataset)

        if data_frame_editor is not None:
            self.df = data_frame_editor(
                self.df, self.path_to_dataset, self.data_readers[list(self.data_readers)[0]]
            )

        # filter the df
        if data_frame_filters is not None:
            if isinstance(data_frame_filters, list):
                for df_filter in data_frame_filters:
                    self.df = df_filter(self.df)
            else:
                self.df = data_frame_filters(self.df)

        # make sure that the indizes starts by 0
        self.df.reset_index(inplace=True, drop=True)

        # create hash for each row
        self.df["__hash_of_row__"] = self.df.apply(lambda row: int(hash(str(row.values)) % 1e17), 1)

        # for preloading the images / labels
        self.preloaded_images = None
        self.preloaded_features = None

    def get_feature_for_all_rows(self, feature_pos=None):
        if self.preloaded_images is not None:
            if feature_pos is None:
                return self.preloaded_features
            else:
                return [row[feature_pos] for row in self.preloaded_features]
        else:
            if feature_pos is None:
                return [
                    feature_loader(self.df, self.path_to_dataset)
                    for feature_loader in self.feature_readers
                ]
            else:
                return self.feature_readers[feature_pos](self.df, self.path_to_dataset)

    def preload(self, batch_size=128):
        self.preloaded_images = {}
        self.preloaded_features = {}
        for i in range(0, len(self.df), batch_size):
            img, feat = self.__load__(self.df.iloc[i : i + batch_size].index)
            for k in img:
                if isinstance(img[k], list):
                    img[k] = np.array(img[k])
                if k in self.preloaded_images:
                    if isinstance(img[k], np.ndarray):
                        self.preloaded_images[k] = np.concatenate(
                            [self.preloaded_images[k], img[k]]
                        )
                    else:
                        self.preloaded_images[k] = torch.cat([self.preloaded_images[k], img[k]], 0)
                else:
                    self.preloaded_images[k] = img[k]

            for k in feat:
                if isinstance(feat[k], list):
                    feat[k] = np.array(feat[k])
                if k in self.preloaded_features:
                    if isinstance(feat[k], np.ndarray):
                        self.preloaded_features[k] = np.concatenate(
                            [self.preloaded_features[k], feat[k]]
                        )
                    else:
                        self.preloaded_features[k] = torch.cat(
                            [self.preloaded_features[k], feat[k]], 0
                        )
                else:
                    self.preloaded_features[k] = feat[k]

    def __load__(self, idx):
        # get asked row
        row = self.df.iloc[idx]

        # load data
        images = {}
        for key, img_loader in self.data_readers.items():
            loaded = img_loader(row, self.path_to_dataset)
            if isinstance(loaded, dict):
                for k in loaded:
                    images[key + "_" + k] = loaded[k]
            else:
                images[key] = loaded
        features = {
            key: feature_loader(row, self.path_to_dataset)
            for key, feature_loader in self.feature_readers.items()
        }

        # pre transform
        images = self.pre_transfs(images, self.__get_hash__(idx))

        return images, features

    def __len__(self):
        return len(self.df)

    def __get_hash__(self, idx):
        # get hashes
        row = self.df.iloc[idx]
        hashes = row["__hash_of_row__"]
        if isinstance(hashes, pd.Series):
            hashes = hashes.values.tolist()
        return hashes

    def __getitem__(self, idx):

        # get asked row
        row = self.df.iloc[idx]

        if self.seed is not None:
            # save random state to reactivate this state after the call
            saved_seed_state = helper.get_random_seed()
            # set seets
            helper.set_random_seed_with_int((1 + self.seed_increment) * int(1e9) + self.seed)
            # increase seed increment to use a new random seed at next call
            self.seed_increment += 1

        if self.preloaded_images is None:
            images, features = self.__load__(idx)
        else:
            images = {}
            for k in self.preloaded_images:
                images[k] = self.preloaded_images[k][idx]
            features = {}
            for k in self.preloaded_features:
                features[k] = self.preloaded_features[k][idx]

        # da transforms
        images = self.da_transfs(images, self.__get_hash__(idx))

        if self.seed is not None:
            # reactivate random state
            helper.set_random_seed(*saved_seed_state)

        dic = {}
        dic.update(images)
        dic.update(features)
        if self.return_data_rows:
            dic["raw"] = row
        return dic



import math
import os
import re

import cv2
import numpy as np
import pandas as pd
from scipy.ndimage.measurements import center_of_mass
from skimage import morphology
from skimage.segmentation import slic


class DataFrameCreator:
    def __init__(self):
        pass

    def __call__(self, path_to_image):
        pass


class CreateDFDummy(DataFrameCreator):
    def __init__(self, num_of_entries=100):
        super().__init__()
        self.num_of_entries = num_of_entries

    def __call__(self, path_to_image):
        df = pd.DataFrame({"id": range(self.num_of_entries)})
        df["fold"] = df.apply(lambda row: row["id"] % 10, 1)
        df["is_even"] = df.apply(lambda row: (row["id"] % 2) == 0, 1)
        return df


class CreateDFFromCSV(DataFrameCreator):
    def __init__(self, path_to_csv, sep=","):
        super().__init__()
        self.path_to_csv = path_to_csv
        self.sep = sep

    def __call__(self, path_to_image):
        if isinstance(self.path_to_csv, list):
            df = pd.DataFrame()
            for path in self.path_to_csv:
                tmp_df = pd.read_csv(path, sep=self.sep)
                tmp_df["__from_csv__"] = path
                df = df.append(tmp_df)
            return df
        else:
            return pd.read_csv(self.path_to_csv, sep=self.sep)


class CreateDFFromFolder(DataFrameCreator):
    def __init__(self, file_of_interest):
        super().__init__()
        self.file_of_interest = file_of_interest

        regex = ""
        tmp = file_of_interest
        self.var_names = []
        counter_unnamed = 0
        while len(tmp) > 0:
            var_start = tmp.find("{")
            var_unnamed_start = tmp.find("(")
            if (
                var_unnamed_start != -1
                and (var_unnamed_start < var_start or var_start == -1)
                and (var_unnamed_start == 0 or tmp[var_unnamed_start - 1] != "\\")
            ):
                regex += tmp[: var_unnamed_start + 1]
                tmp = tmp[var_unnamed_start + 1 :]
                self.var_names.append("unnamed_" + str(counter_unnamed))
                counter_unnamed += 1
            elif var_start == -1:  # No Variable could be found
                regex += tmp
                tmp = ""
            else:
                # ignore this var start
                if var_start > 0 and tmp[var_start - 1] == "\\":
                    regex += tmp[:var_start]
                    tmp = tmp[var_start:]
                else:
                    regex += tmp[:var_start]
                    tmp = tmp[var_start:]
                    var_end = tmp.find("}")
                    self.var_names.append(tmp[1:var_end])
                    # self.regex += '(?=.*)'
                    # regex += '(^[^/]*)'
                    regex += "(.*)"
                    tmp = tmp[var_end + 1 :]
        self.regex = re.compile(regex)

    def __call__(self, path_to_image):
        rows = {name: [] for name in self.var_names}
        rows["__path__"] = []
        for root, _, files in os.walk(path_to_image):
            for file in files:
                path = os.path.join(root, file)[len(path_to_image) :]
                result = self.regex.search(path)
                if result is not None:
                    path_as_variable = False
                    row = {}
                    for vname, var in zip(self.var_names, result.groups()):
                        row[vname] = var
                        if var.find("/") != -1:
                            path_as_variable = True
                    if path_as_variable is False:
                        for k in row:
                            rows[k].append(row[k])
                        rows["__path__"].append(path)

        return pd.DataFrame(rows)


class DataFrameEditor:
    def __init__(self):
        pass

    def __call__(self, df, path_to_image):
        pass


class EditDFImageGrid(DataFrameEditor):
    def __init__(self, return_image_size, perc_overlapping=0.0, input_im_size=None):
        super().__init__()
        self.return_image_size = return_image_size
        self.perc_overlapping = perc_overlapping
        self.input_im_size = input_im_size

    def get_points(self, im_size1d):
        im_size_center = im_size1d - self.return_image_size
        shift = self.return_image_size * (1 - self.perc_overlapping)
        locations = np.arange(0, im_size_center + 0.0001, shift)

        return np.array(locations + (im_size1d - locations[-1]) / 2, dtype=int)

    def __call__(self, df, path_to_image, reader):
        if self.input_im_size is not None:
            input_im_size = self.input_im_size

        rows = {name: [] for name in list(df)}
        rows["__x__"] = []
        rows["__y__"] = []
        for _, row in df.iterrows():
            if self.input_im_size is None:
                input_im_size = reader.get_image_size(row, path_to_image)
            xs = self.get_points(input_im_size[0])
            ys = self.get_points(input_im_size[1])

            for x in xs:
                for y in ys:
                    row = dict(row)
                    row["__x__"] = x
                    row["__y__"] = y

                    for k in row:
                        rows[k].append(row[k])

        return pd.DataFrame(rows)


# an example function using intensity thresholding and morphological
# operations to obtain tissue regions
def simple_get_mask(rgb):
    gray = cv2.cvtColor(rgb, cv2.COLOR_RGB2GRAY)
    _, mask = cv2.threshold(gray, 0, 255, cv2.THRESH_OTSU)  # threshold using OTSU method
    mask = morphology.remove_small_objects(mask == 0, min_size=100, connectivity=2)
    mask = morphology.remove_small_holes(mask, area_threshold=100)
    mask = morphology.binary_dilation(mask, morphology.disk(5))
    return mask


class EditDFWSIGrid(DataFrameEditor):
    def __init__(self, return_image_size, perc_overlapping=0.0):
        super().__init__()
        self.return_image_size = return_image_size
        self.perc_overlapping = perc_overlapping

    def __get_coordinates__(self, thumbnail, thumbnail_mpp, return_mpp):
        print(thumbnail.shape)
        # see: https://colab.research.google.com/github/TIA-Lab
        # /tiatoolbox/blob/master/examples/example_wsiread.ipynb#scrollTo=MZ_yqoGJ_-6i
        wsi_thumb_mask = simple_get_mask(thumbnail)

        lores_patch_size = int(self.return_image_size / (thumbnail_mpp / return_mpp))
        nr_expected_rois = math.ceil(np.sum(wsi_thumb_mask) / (lores_patch_size ** 2))

        print(lores_patch_size, nr_expected_rois)

        wsi_rois_mask = slic(
            thumbnail, mask=wsi_thumb_mask, n_segments=nr_expected_rois, compactness=1000, sigma=1
        )

        lores_rois_center = center_of_mass(
            wsi_rois_mask, labels=wsi_rois_mask, index=np.unique(wsi_rois_mask)[1:]
        )
        lores_rois_center = np.array(lores_rois_center)  # coordinates is Y, X
        lores_rois_center = lores_rois_center.astype(np.int32)
        selected_indices = wsi_thumb_mask[lores_rois_center[:, 0], lores_rois_center[:, 1]]
        lores_rois_center = lores_rois_center[selected_indices]

        return lores_rois_center * (thumbnail_mpp / return_mpp)

    def __call__(self, df, path_to_image, reader):

        rows = {name: [] for name in list(df)}
        rows["__x__"] = []
        rows["__y__"] = []
        for _, row in df.iterrows():

            thumbnail = reader.read_thumbnail(row, path_to_image, 50)
            coordinates = self.__get_coordinates__(thumbnail, 50, reader.mpp)

            for c in coordinates:
                row = dict(row)
                row["__x__"] = int(c[0])
                row["__y__"] = int(c[1])

                for k in row:
                    rows[k].append(row[k])

        return pd.DataFrame(rows)



import copy

import numpy as np


class TransformerAdapter:
    def __init__(self):
        self.is_callable_with_batches = False
        self.is_splitter = False

    def __call__(self, img_or_imgs):
        pass


class RandomPrinterAdapter(TransformerAdapter):
    def __init__(self, message="RandomPrinterAdapter:"):
        super().__init__()
        self.message = message
        self.is_callable_with_batches = True
        self.is_splitter = False

    def __call__(self, img_or_imgs):
        print(self.message, np.random.random())
        return img_or_imgs


class CopyAdapter(TransformerAdapter):
    def __init__(self, num_of_copies=2):
        super().__init__()
        self.is_callable_with_batches = True
        self.is_splitter = True
        self.num_of_copies = num_of_copies

    def __call__(self, img_or_imgs):
        result = {}
        for i in range(self.num_of_copies):
            # TODO
            # result.append(img_or_imgs.clone())
            result[str(i)] = copy.copy(img_or_imgs)
        return result



import random


class DataFrameFilter:
    def __init__(self, mode):
        self.mode = mode

    def __call__(self, df, indices):
        if self.mode == 0:
            return df.iloc[indices]
        elif self.mode == 1:
            return df.loc[~df.index.isin(indices)]
        else:
            return df.iloc[indices], df.loc[~df.index.isin(indices)]


class RandomFilter(DataFrameFilter):
    def __init__(self, percentage_to_filter, mode=0, seed=None):
        super().__init__(mode)
        self.percentage_to_filter = percentage_to_filter
        self.seed = seed

    def __call__(self, df):
        r = random.Random(self.seed)
        indices = r.sample(list(df.index), int(len(df) * self.percentage_to_filter))
        return super().__call__(df, indices)


class LambdaFilter(DataFrameFilter):
    def __init__(self, lambda_function, mode=0):
        super().__init__(mode)
        self.lambda_function = lambda_function

    def __call__(self, df):
        return super().__call__(df, df[df.apply(self.lambda_function, 1)].index)


class ColumnFilter(DataFrameFilter):
    def __init__(self, name_of_column, allowed_values, mode=0):
        super().__init__(mode)
        self.name_of_column = name_of_column
        self.allowed_values = allowed_values

    def __call__(self, df):
        if isinstance(self.allowed_values, list):
            return super().__call__(df, df[df[self.name_of_column].isin(self.allowed_values)].index)
        else:
            return super().__call__(df, df[df[self.name_of_column] == self.allowed_values].index)
